@@ -17,7 +17,7 @@ conditions.
 #include "nc4dispatch.h"
 #include "ncdispatch.h"
 
-#ifdef USE_PNETCDF
+#if 0 /*def USE_PNETCDF*/
 #include <pnetcdf.h>
 #endif
 
@@ -27,9 +27,9 @@ int nc4typelen(nc_type type);
    info. Always locate the attribute by name, never by attnum.
    The mem_type is ignored if data=NULL. */
 int
-nc4_get_att(int ncid, NC *nc, int varid, const char *name, 
-	    nc_type *xtype, nc_type mem_type, size_t *lenp, 
-	    int *attnum, int is_long, void *data) 
+nc4_get_att(int ncid, NC *nc, int varid, const char *name,
+	    nc_type *xtype, nc_type mem_type, size_t *lenp,
+	    int *attnum, int is_long, void *data)
 {
    NC_GRP_INFO_T *grp;
    NC_HDF5_FILE_INFO_T *h5;
@@ -38,7 +38,7 @@ nc4_get_att(int ncid, NC *nc, int varid, const char *name,
 
    int need_to_convert = 0;
    int range_error = NC_NOERR;
-   void *bufr = NULL; 
+   void *bufr = NULL;
    size_t type_size;
    char norm_name[NC_MAX_NAME + 1];
    int i;
@@ -48,73 +48,73 @@ nc4_get_att(int ncid, NC *nc, int varid, const char *name,
       my_attnum = *attnum;
    assert(nc && NC4_DATA(nc));
 
-   LOG((3, "nc4_get_att: ncid 0x%x varid %d name %s attnum %d mem_type %d", 
-	ncid, varid, name, my_attnum, mem_type));
+   LOG((3, "%s: ncid 0x%x varid %d name %s attnum %d mem_type %d",
+	__func__, ncid, varid, name, my_attnum, mem_type));
 
    /* Find info for this file and group, and set pointer to each. */
    h5 = NC4_DATA(nc);
    if (!(grp = nc4_rec_find_grp(h5->root_grp, (ncid & GRP_ID_MASK))))
-      return NC_EBADGRPID;      
+      BAIL(NC_EBADGRPID);
 
    /* Normalize name. */
    if ((retval = nc4_normalize_name(name, norm_name)))
-      return retval;
+      BAIL(retval);
 
    /* Find the attribute, if it exists. If we don't find it, we are
       major failures. */
    if ((retval = nc4_find_grp_att(grp, varid, norm_name, my_attnum, &att)))
-      return retval;
-   
+      BAIL(retval);
+
    /* If mem_type is NC_NAT, it means we want to use the attribute's
     * file type as the mem type as well. */
    if (mem_type == NC_NAT)
-      mem_type = att->xtype;
+      mem_type = att->nc_typeid;
 
    /* If the attribute is NC_CHAR, and the mem_type isn't, or vice
     * versa, that's a freakish attempt to convert text to
     * numbers. Some pervert out there is trying to pull a fast one!
     * Send him an NC_ECHAR error...*/
    if (data && att->len &&
-       ((att->xtype == NC_CHAR && mem_type != NC_CHAR) ||
-	(att->xtype != NC_CHAR && mem_type == NC_CHAR)))
-      return NC_ECHAR; /* take that, you freak! */
+       ((att->nc_typeid == NC_CHAR && mem_type != NC_CHAR) ||
+	(att->nc_typeid != NC_CHAR && mem_type == NC_CHAR)))
+      BAIL(NC_ECHAR); /* take that, you freak! */
 
    /* Copy the info. */
    if (lenp)
       *lenp = att->len;
    if (xtype)
-      *xtype = att->xtype;
+      *xtype = att->nc_typeid;
    if (attnum)
       *attnum = att->attnum;
 
    /* Zero len attributes are easy to read! */
    if (!att->len)
-      return NC_NOERR;
+      BAIL(NC_NOERR);
 
    /* Later on, we will need to know the size of this type. */
    if ((retval = nc4_get_typelen_mem(h5, mem_type, is_long, &type_size)))
-      return retval;
+      BAIL(retval);
 
    /* We may have to convert data. Treat NC_CHAR the same as
     * NC_UBYTE. If the mem_type is NAT, don't try any conversion - use
     * the attribute's type. */
-   if (data && att->len && mem_type != att->xtype &&
+   if (data && att->len && mem_type != att->nc_typeid &&
        mem_type != NC_NAT &&
-       !(mem_type == NC_CHAR && 
-	 (att->xtype == NC_UBYTE || att->xtype == NC_BYTE)))
+       !(mem_type == NC_CHAR &&
+	 (att->nc_typeid == NC_UBYTE || att->nc_typeid == NC_BYTE)))
    {
-      need_to_convert++;
       if (!(bufr = malloc((size_t)(att->len * type_size))))
-	 return NC_ENOMEM;
-      if ((retval = nc4_convert_type(att->data, bufr, att->xtype, 
-				     mem_type, (size_t)att->len, &range_error, 
+	 BAIL(NC_ENOMEM);
+      need_to_convert++;
+      if ((retval = nc4_convert_type(att->data, bufr, att->nc_typeid,
+				     mem_type, (size_t)att->len, &range_error,
 				     NULL, (h5->cmode & NC_CLASSIC_MODEL), 0, is_long)))
 	 BAIL(retval);
 
       /* For strict netcdf-3 rules, ignore erange errors between UBYTE
        * and BYTE types. */
       if ((h5->cmode & NC_CLASSIC_MODEL) &&
-	  (att->xtype == NC_UBYTE || att->xtype == NC_BYTE) &&
+	  (att->nc_typeid == NC_UBYTE || att->nc_typeid == NC_BYTE) &&
 	  (mem_type == NC_UBYTE || mem_type == NC_BYTE) &&
 	  range_error)
 	 range_error = 0;
@@ -125,20 +125,25 @@ nc4_get_att(int ncid, NC *nc, int varid, const char *name,
    }
 
    /* If the caller wants data, copy it for him. If he hasn't
-      allocated enough memory for it, he will burn in segmantation
+      allocated enough memory for it, he will burn in segmentation
       fault hell, writhing with the agony of undiscovered memory
       bugs! */
    if (data)
    {
       if (att->vldata)
       {
-	 size_t base_typelen = type_size;
+	 size_t base_typelen;
 	 hvl_t *vldest = data;
 	 NC_TYPE_INFO_T *type;
-	 if ((retval = nc4_find_type(h5, att->xtype, &type))) {
-	   if(bufr) free(bufr);
-	   return retval;
-	 }
+
+         /* Get the type object for the attribute's type */
+	 if ((retval = nc4_find_type(h5, att->nc_typeid, &type)))
+	   BAIL(retval);
+
+         /* Retrieve the size of the base type */
+         if ((retval = nc4_get_typelen_mem(h5, type->u.v.base_nc_typeid, 0, &base_typelen)))
+            BAIL(retval);
+
 	 for (i = 0; i < att->len; i++)
 	 {
 	    vldest[i].len = att->vldata[i].len;
@@ -151,18 +156,24 @@ nc4_get_att(int ncid, NC *nc, int varid, const char *name,
       {
 	 for (i = 0; i < att->len; i++)
 	 {
-	    if (!(((char **)data)[i] = malloc(strlen(att->stdata[i]) + 1)))
-	       BAIL(NC_ENOMEM);
-	    strcpy(((char **)data)[i], att->stdata[i]);
+            /* Check for NULL pointer for string (valid in HDF5) */
+            if(att->stdata[i])
+            {
+                if (!(((char **)data)[i] = strdup(att->stdata[i])))
+                   BAIL(NC_ENOMEM);
+            }
+            else
+                ((char **)data)[i] = att->stdata[i];
 	 }
       }
       else
       {
 	 /* For long types, we need to handle this special... */
-	 if (is_long && att->xtype == NC_INT)
+	 if (is_long && att->nc_typeid == NC_INT)
 	 {
 	    long *lp = data;
 	    int *ip = bufr;
+
 	    for (i = 0; i < att->len; i++)
 	       *lp++ = *ip++;
 	 }
@@ -172,33 +183,31 @@ nc4_get_att(int ncid, NC *nc, int varid, const char *name,
    }
 
  exit:
-   if (need_to_convert) free(bufr);
-   if (retval)
-      return retval;
+   if (need_to_convert)
+      free(bufr);
    if (range_error)
-      return NC_ERANGE;
-   return NC_NOERR;
+      retval = NC_ERANGE;
+   return retval;
 }
 
 /* Put attribute metadata into our global metadata. */
-int
-nc4_put_att(int ncid, NC *nc, int varid, const char *name, 
-	    nc_type file_type, nc_type mem_type, size_t len, int is_long, 
+static int
+nc4_put_att(int ncid, NC *nc, int varid, const char *name,
+	    nc_type file_type, nc_type mem_type, size_t len, int is_long,
 	    const void *data)
 {
-   NC_GRP_INFO_T *grp; 
+   NC_GRP_INFO_T *grp;
    NC_HDF5_FILE_INFO_T *h5;
    NC_VAR_INFO_T *var = NULL;
-   NC_ATT_INFO_T *att, **attlist = NULL, *varatt;
-   NC_TYPE_INFO_T *type = NULL;
+   NC_ATT_INFO_T *att, **attlist = NULL;
    char norm_name[NC_MAX_NAME + 1];
-   int new_att = 0;
+   nc_bool_t new_att = NC_FALSE;
    int retval = NC_NOERR, range_error = 0;
    size_t type_size;
    int i;
    int res;
 
-   if (!name) 
+   if (!name)
       return NC_EBADNAME;
    assert(nc && NC4_DATA(nc));
 
@@ -213,7 +222,7 @@ nc4_put_att(int ncid, NC *nc, int varid, const char *name,
    /* Find info for this file and group, and set pointer to each. */
    h5 = NC4_DATA(nc);
    if (!(grp = nc4_rec_find_grp(h5->root_grp, (ncid & GRP_ID_MASK))))
-      return NC_EBADGRPID;      
+      return NC_EBADGRPID;
 
    /* If the file is read-only, return an error. */
    if (h5->no_write)
@@ -228,7 +237,7 @@ nc4_put_att(int ncid, NC *nc, int varid, const char *name,
       attlist = &grp->att;
    else
    {
-      for (var = grp->var; var; var = var->next)
+      for (var = grp->var; var; var = var->l.next)
 	 if (var->varid == varid)
 	 {
 	    attlist = &var->att;
@@ -237,7 +246,7 @@ nc4_put_att(int ncid, NC *nc, int varid, const char *name,
       if (!var)
 	 return NC_ENOTVAR;
    }
-   for (att = *attlist; att; att = att->next)
+   for (att = *attlist; att; att = att->l.next)
       if (!strcmp(att->name, norm_name))
 	 break;
 
@@ -251,14 +260,14 @@ nc4_put_att(int ncid, NC *nc, int varid, const char *name,
 	 if ((retval = NC4_redef(ncid)))
 	    BAIL(retval);
       }
-      new_att++;
+      new_att = NC_TRUE;
    }
    else
    {
       /* For an existing att, if we're not in define mode, the len
 	 must not be greater than the existing len for classic model. */
-      if (!(h5->flags & NC_INDEF) && 
-	  len * nc4typelen(file_type) > (size_t)att->len * nc4typelen(att->xtype))
+      if (!(h5->flags & NC_INDEF) &&
+	  len * nc4typelen(file_type) > (size_t)att->len * nc4typelen(att->nc_typeid))
       {
 	 if (h5->cmode & NC_CLASSIC_MODEL)
 	    return NC_EINDEFINE;
@@ -272,14 +281,12 @@ nc4_put_att(int ncid, NC *nc, int varid, const char *name,
       return NC_EBADTYPE;
 
    /* Get information about this type. */
-   if ((retval = nc4_find_type(h5, file_type, &type)))
-      return retval;
    if ((retval = nc4_get_typelen_mem(h5, file_type, is_long, &type_size)))
       return retval;
 
    /* No character conversions are allowed. */
-   if (file_type != mem_type && 
-       (file_type == NC_CHAR || mem_type == NC_CHAR || 
+   if (file_type != mem_type &&
+       (file_type == NC_CHAR || mem_type == NC_CHAR ||
 	file_type == NC_STRING || mem_type == NC_STRING))
       return NC_ECHAR;
 
@@ -293,40 +300,50 @@ nc4_put_att(int ncid, NC *nc, int varid, const char *name,
    if (new_att)
    {
       LOG((3, "adding attribute %s to the list...", norm_name));
-      if ((res = nc4_att_list_add(attlist)))
+      if ((res = nc4_att_list_add(attlist, &att)))
 	 BAIL (res);
-      /* Find this att's entry in the list (the last one). */
-      for (att=*attlist; att->next; att=att->next)
-	 ;
+      if (!(att->name = strdup(norm_name)))
+         return NC_ENOMEM;
    }
 
    /* Now fill in the metadata. */
-   att->dirty++;
-   if (att->name)
-      free(att->name);
-   if (!(att->name = malloc((strlen(norm_name) + 1) * sizeof(char))))
-      return NC_ENOMEM;
-   strcpy(att->name, norm_name);
-   att->xtype = file_type;
+   att->dirty = NC_TRUE;
+   att->nc_typeid = file_type;
+
+   /* If this att has vlen or string data, release it before we lose the length value. */
+   if (att->stdata)
+   {
+      for (i = 0; i < att->len; i++)
+         if(att->stdata[i])
+	    free(att->stdata[i]);
+      free(att->stdata);
+      att->stdata = NULL;
+   }
+   if (att->vldata)
+   {
+      for (i = 0; i < att->len; i++)
+	 nc_free_vlen(&att->vldata[i]);
+      free(att->vldata);
+      att->vldata = NULL;
+   }
+
    att->len = len;
-   if (att->prev)
-      att->attnum = att->prev->attnum + 1;
+   if (att->l.prev)
+      att->attnum = ((NC_ATT_INFO_T *)att->l.prev)->attnum + 1;
    else
       att->attnum = 0;
-   if (type)
-      att->class = type->class;
 
    /* If this is the _FillValue attribute, then we will also have to
-    * copy the value to the fll_vlue pointer of the NC_VAR_INFO_T
+    * copy the value to the fill_vlue pointer of the NC_VAR_INFO_T
     * struct for this var. (But ignore a global _FillValue
     * attribute). */
    if (!strcmp(att->name, _FillValue) && varid != NC_GLOBAL)
    {
-      NC_TYPE_INFO_T *type_info;
+      NC_ATT_INFO_T *varatt;
       int size;
 
       /* Fill value must be same type and have exactly one value */
-      if (att->xtype != var->xtype)
+      if (att->nc_typeid != var->type_info->nc_typeid)
 	 return NC_EBADTYPE;
       if (att->len != 1)
 	 return NC_EINVAL;
@@ -336,124 +353,154 @@ nc4_put_att(int ncid, NC *nc, int varid, const char *name,
 	 return NC_ELATEFILL;
 
       /* If fill value hasn't been set, allocate space. Of course,
-       * vlens have to be differnt... */
-      if ((retval = nc4_get_typelen_mem(grp->nc4_info, var->xtype, 0, 
+       * vlens have to be different... */
+      if ((retval = nc4_get_typelen_mem(grp->nc4_info, var->type_info->nc_typeid, 0,
 					&type_size)))
 	 return retval;
-      if ((retval = nc4_find_type(grp->nc4_info, var->xtype, &type_info)))
-	 BAIL(retval);
-      
+
       /* Already set a fill value? Now I'll have to free the old
        * one. Make up your damn mind, would you? */
       if (var->fill_value)
       {
-	 if (type_info && type_info->class == NC_VLEN)
-	    if ((retval = nc_free_vlen(var->fill_value)))
-	       return retval;
+         if (var->type_info->nc_type_class == NC_VLEN)
+         {
+            if ((retval = nc_free_vlen(var->fill_value)))
+               return retval;
+         }
+         else if (var->type_info->nc_type_class == NC_STRING)
+         {
+            if (*(char **)var->fill_value)
+               free(*(char **)var->fill_value);
+         }
 	 free(var->fill_value);
       }
 
       /* Allocate space for the fill value. */
-      if (type_info && type_info->class == NC_VLEN)
+      if (var->type_info->nc_type_class == NC_VLEN)
 	 size = sizeof(hvl_t);
-      else if (var->xtype == NC_STRING)
+      else if (var->type_info->nc_type_class == NC_STRING)
 	 size = sizeof(char *);
       else
 	 size = type_size;
 
-      /* 	 size = strlen(*(char **)data) + 1; */
-      if (!(var->fill_value = malloc(size)))
+      if (!(var->fill_value = calloc(1, size)))
 	 return NC_ENOMEM;
 
       /* Copy the fill_value. */
       LOG((4, "Copying fill value into metadata for variable %s", var->name));
-      if (type_info && type_info->class == NC_VLEN)
+      if (var->type_info->nc_type_class == NC_VLEN)
       {
 	 nc_vlen_t *in_vlen = (nc_vlen_t *)data, *fv_vlen = (nc_vlen_t *)(var->fill_value);
+
 	 fv_vlen->len = in_vlen->len;
 	 if (!(fv_vlen->p = malloc(size * in_vlen->len)))
 	    return NC_ENOMEM;
 	 memcpy(fv_vlen->p, in_vlen->p, in_vlen->len * size);
       }
-      else if (var->xtype == NC_STRING)
+      else if (var->type_info->nc_type_class == NC_STRING)
       {
-	 if (!(*(char **)var->fill_value = malloc(strlen(*(char **)data) + 1)))
-	    return NC_ENOMEM;
-	 strcpy(*(char **)(var->fill_value), *(char **)data);
+         if(NULL != (*(char **)data))
+         {
+            if (!(*(char **)(var->fill_value) = malloc(strlen(*(char **)data) + 1)))
+               return NC_ENOMEM;
+            strcpy(*(char **)var->fill_value, *(char **)data);
+         }
+         else
+            *(char **)var->fill_value = NULL;
       }
       else
 	 memcpy(var->fill_value, data, type_size);
-
-      /* Mark the var and all its atts as dirty, so they get
-       * rewritten. */
-      var->dirty++;
-      for (varatt = var->att; varatt; varatt = varatt->next)
-	 varatt->dirty++;
+      
+      /* Indicate that the fill value was changed, if the variable has already
+       * been created in the file, so the dataset gets deleted and re-created. */
+      if (var->created)
+         var->fill_val_changed = NC_TRUE;
    }
 
    /* Copy the attribute data, if there is any. VLENs and string
     * arrays have to be handled specially. */
-   if (type && type->class == NC_VLEN && data && att->len)
+   if(att->len)
    {
-      const hvl_t *vldata1;
+      nc_type type_class;    /* Class of attribute's type */
 
-      vldata1 = data;
-      if (!(att->vldata = malloc(att->len * sizeof(hvl_t))))
-	 BAIL(NC_ENOMEM);	 
-      for (i = 0; i < att->len; i++)
+      /* Get class for this type. */
+      if ((retval = nc4_get_typeclass(h5, file_type, &type_class)))
+         return retval;
+
+      assert(data);
+      if (type_class == NC_VLEN)
       {
-	 att->vldata[i].len = vldata1[i].len;
-	 if (!(att->vldata[i].p = malloc(type_size * att->vldata[i].len)))
-	    BAIL(NC_ENOMEM);
-	 memcpy(att->vldata[i].p, vldata1[i].p, type_size * att->vldata[i].len);
+         const hvl_t *vldata1;
+	 NC_TYPE_INFO_T *type;
+	 size_t base_typelen;
+
+         /* Get the type object for the attribute's type */
+	 if ((retval = nc4_find_type(h5, file_type, &type)))
+	   BAIL(retval);
+
+         /* Retrieve the size of the base type */
+         if ((retval = nc4_get_typelen_mem(h5, type->u.v.base_nc_typeid, 0, &base_typelen)))
+            BAIL(retval);
+
+         vldata1 = data;
+         if (!(att->vldata = (nc_vlen_t*)malloc(att->len * sizeof(hvl_t))))
+            BAIL(NC_ENOMEM);
+         for (i = 0; i < att->len; i++)
+         {
+            att->vldata[i].len = vldata1[i].len;
+            if (!(att->vldata[i].p = malloc(base_typelen * att->vldata[i].len)))
+               BAIL(NC_ENOMEM);
+            memcpy(att->vldata[i].p, vldata1[i].p, base_typelen * att->vldata[i].len);
+         }
+      }
+      else if (type_class == NC_STRING)
+      {
+         LOG((4, "copying array of NC_STRING"));
+         if (!(att->stdata = malloc(sizeof(char *) * att->len)))
+            BAIL(NC_ENOMEM);
+         for (i = 0; i < att->len; i++)
+         {
+            if(NULL != ((char **)data)[i]) {
+               LOG((5, "copying string %d of size %d", i, strlen(((char **)data)[i]) + 1));
+               if (!(att->stdata[i] = strdup(((char **)data)[i])))
+                  BAIL(NC_ENOMEM);
+           }
+           else
+               att->stdata[i] = ((char **)data)[i];
+         }
+      }
+      else
+      {
+         /* [Re]allocate memory for the attribute data */
+         if (!new_att)
+            free (att->data);
+         if (!(att->data = malloc(att->len * type_size)))
+            BAIL(NC_ENOMEM);
+
+         /* Just copy the data, for non-atomic types */
+         if (type_class == NC_OPAQUE || type_class == NC_COMPOUND || type_class == NC_ENUM)
+            memcpy(att->data, data, len * type_size);
+         else
+         {
+            /* Data types are like religions, in that one can convert.  */
+            if ((retval = nc4_convert_type(data, att->data, mem_type, file_type,
+                                           len, &range_error, NULL,
+                                           (h5->cmode & NC_CLASSIC_MODEL), is_long, 0)))
+               BAIL(retval);
+         }
       }
    }
-   else if (file_type == NC_STRING && data && att->len)
-   {
-      LOG((4, "copying array of NC_STRING"));
-      if (!(att->stdata = malloc(sizeof(char *) * att->len)))
-	 BAIL(NC_ENOMEM);	 
-      for (i = 0; i < att->len; i++)
-      {
-	 LOG((5, "copying string %d of size %d", i, strlen(((char **)data)[i]) + 1));
-	 if (!(att->stdata[i] = malloc(strlen(((char **)data)[i]) + 1)))
-	    BAIL(NC_ENOMEM);
-	 strcpy(att->stdata[i], ((char **)data)[i]);
-      }
-   }
-   else
-   {
-      /* Data types are like religions, in that one can convert.  */
-      if (att->len)
-      {
-	 if (!new_att)
-	    free (att->data);
-	 if (!(att->data = malloc(att->len * type_size)))
-	    BAIL(NC_ENOMEM);
-	 if (type)
-	 {
-	    /* Just copy the data... */
-	    if (type->class == NC_OPAQUE || type->class == NC_COMPOUND || type->class == NC_ENUM)
-	       memcpy(att->data, data, len * type_size);
-	    else
-	       LOG((0, "nc4_put_att: unknown type."));
-	 }
-	 else
-	 {
-	    if ((retval = nc4_convert_type(data, att->data, mem_type, file_type, 
-					   len, &range_error, NULL, 
-					   (h5->cmode & NC_CLASSIC_MODEL), is_long, 0)))
-	       BAIL(retval);
-	 }
-      }
-   }
-   att->dirty = 1;
-   att->created = 0;
+   att->dirty = NC_TRUE;
+   att->created = NC_FALSE;
+
+   /* Mark attributes on variable dirty, so they get written */
+   if(var)
+       var->attr_dirty = NC_TRUE;
 
  exit:
    /* If there was an error return it, otherwise return any potential
       range error value. If none, return NC_NOERR as usual.*/
-   if (retval)      
+   if (retval)
       return retval;
    if (range_error)
       return NC_ERANGE;
@@ -478,12 +525,12 @@ NC4_inq_att(int ncid, int varid, const char *name, nc_type *xtypep, size_t *lenp
    h5 = NC4_DATA(nc);
    assert(h5);
 
-#ifdef USE_PNETCDF
+#if 0 /*def USE_PNETCDF*/
    /* Take care of files created/opened with parallel-netcdf library. */
    if (h5->pnetcdf_file)
    {
       MPI_Offset mpi_len;
-      int ret = ncmpi_inq_att(nc->int_ncid, varid, name, xtypep, &mpi_len);	
+      int ret = ncmpi_inq_att(nc->int_ncid, varid, name, xtypep, &mpi_len);
       if (ret != NC_NOERR)
 	 return ret;
       if (lenp)
@@ -496,7 +543,7 @@ NC4_inq_att(int ncid, int varid, const char *name, nc_type *xtypep, size_t *lenp
 }
 
 /* Learn an attnum, given a name. */
-int 
+int
 NC4_inq_attid(int ncid, int varid, const char *name, int *attnump)
 {
    NC *nc;
@@ -512,14 +559,14 @@ NC4_inq_attid(int ncid, int varid, const char *name, int *attnump)
    h5 = NC4_DATA(nc);
    assert(h5);
 
-#ifdef USE_PNETCDF
+#if 0 /*def USE_PNETCDF*/
    /* Take care of files created/opened with parallel-netcdf library. */
    if (h5->pnetcdf_file)
       return ncmpi_inq_attid(nc->int_ncid, varid, name, attnump);
 #endif /* USE_PNETCDF */
 
    /* Handle netcdf-4 files. */
-   return nc4_get_att(ncid, nc, varid, name, NULL, NC_UBYTE, 
+   return nc4_get_att(ncid, nc, varid, name, NULL, NC_UBYTE,
 		      NULL, attnump, 0, NULL);
 }
 
@@ -533,7 +580,7 @@ NC4_inq_attname(int ncid, int varid, int attnum, char *name)
    NC_HDF5_FILE_INFO_T *h5;
    int retval = NC_NOERR;
 
-   LOG((2, "nc_inq_attname: ncid 0x%x varid %d attnum %d", 
+   LOG((2, "nc_inq_attname: ncid 0x%x varid %d attnum %d",
 	ncid, varid, attnum));
 
    /* Find metadata. */
@@ -544,7 +591,7 @@ NC4_inq_attname(int ncid, int varid, int attnum, char *name)
    h5 = NC4_DATA(nc);
    assert(h5);
 
-#ifdef USE_PNETCDF
+#if 0 /*def USE_PNETCDF*/
    /* Take care of files created/opened with parallel-netcdf library. */
    if (h5->pnetcdf_file)
       return ncmpi_inq_attname(nc->int_ncid, varid, attnum, name);
@@ -564,11 +611,11 @@ NC4_inq_attname(int ncid, int varid, int attnum, char *name)
 /* I think all atts should be named the exact same thing, to avoid
    confusion! */
 int
-NC4_rename_att(int ncid, int varid, const char *name, 
+NC4_rename_att(int ncid, int varid, const char *name,
 	      const char *newname)
 {
    NC *nc;
-   NC_GRP_INFO_T *grp; 
+   NC_GRP_INFO_T *grp;
    NC_HDF5_FILE_INFO_T *h5;
    NC_VAR_INFO_T *var;
    NC_ATT_INFO_T *att, *list;
@@ -596,7 +643,7 @@ NC4_rename_att(int ncid, int varid, const char *name,
    if (h5->no_write)
      return NC_EPERM;
 
-#ifdef USE_PNETCDF
+#if 0 /*def USE_PNETCDF*/
    /* Take care of files created/opened with parallel-netcdf library. */
    if (h5->pnetcdf_file)
       return ncmpi_rename_att(nc->int_ncid, varid, name, newname);
@@ -613,7 +660,7 @@ NC4_rename_att(int ncid, int varid, const char *name,
    }
    else
    {
-      for (var = grp->var; var; var = var->next)
+      for (var = grp->var; var; var = var->l.next)
 	 if (var->varid == varid)
 	 {
 	    list = var->att;
@@ -622,14 +669,14 @@ NC4_rename_att(int ncid, int varid, const char *name,
       if (!var)
 	 return NC_ENOTVAR;
    }
-   for (att = list; att; att = att->next)
+   for (att = list; att; att = att->l.next)
       if (!strncmp(att->name, norm_newname, NC_MAX_NAME))
 	 return NC_ENAMEINUSE;
 
    /* Normalize name and find the attribute. */
    if ((retval = nc4_normalize_name(name, norm_name)))
       return retval;
-   for (att = list; att; att = att->next)
+   for (att = list; att; att = att->l.next)
       if (!strncmp(att->name, norm_name, NC_MAX_NAME))
 	 break;
    if (!att)
@@ -656,7 +703,7 @@ NC4_rename_att(int ncid, int varid, const char *name,
          if (H5Adelete(datasetid, att->name) < 0)
 	    return NC_EHDFERR;
       }
-      att->created = 0;
+      att->created = NC_FALSE;
    }
 
    /* Copy the new name into our metadata. */
@@ -664,21 +711,21 @@ NC4_rename_att(int ncid, int varid, const char *name,
    if (!(att->name = malloc((strlen(norm_newname) + 1) * sizeof(char))))
       return NC_ENOMEM;
    strcpy(att->name, norm_newname);
-   att->dirty = 1;
+   att->dirty = NC_TRUE;
 
    return retval;
 }
 
 /* Delete an att. Rub it out. Push the button on it. Liquidate
    it. Bump it off. Take it for a one-way ride. Terminate it. Drop the
-   bomb on it. You get the idea. 
-   Ed Hartnett, 10/1/3 
+   bomb on it. You get the idea.
+   Ed Hartnett, 10/1/3
 */
 int
 NC4_del_att(int ncid, int varid, const char *name)
 {
    NC *nc;
-   NC_GRP_INFO_T *grp; 
+   NC_GRP_INFO_T *grp;
    NC_HDF5_FILE_INFO_T *h5;
    NC_ATT_INFO_T *att, *natt;
    NC_VAR_INFO_T *var;
@@ -691,7 +738,7 @@ NC4_del_att(int ncid, int varid, const char *name)
 
    LOG((2, "nc_del_att: ncid 0x%x varid %d name %s",
 	ncid, varid, name));
-   
+
    /* Find metadata for this file. */
    if ((retval = nc4_find_nc_grp_h5(ncid, &nc, &grp, &h5)))
       return retval;
@@ -702,10 +749,10 @@ NC4_del_att(int ncid, int varid, const char *name)
    if (h5->no_write)
       return NC_EPERM;
 
-#ifdef USE_PNETCDF
+#if 0 /*def USE_PNETCDF*/
    /* Take care of files created/opened with parallel-netcdf library. */
    if (h5->pnetcdf_file)
-      return ncmpi_del_att(nc->int_ncid, varid, name);   
+      return ncmpi_del_att(nc->int_ncid, varid, name);
 #endif /* USE_PNETCDF */
 
    /* If it's not in define mode, forget it. */
@@ -726,7 +773,7 @@ NC4_del_att(int ncid, int varid, const char *name)
    }
    else
    {
-      for(var = grp->var; var; var = var->next)
+      for(var = grp->var; var; var = var->l.next)
       {
 	 if (var->varid == varid)
 	 {
@@ -737,42 +784,44 @@ NC4_del_att(int ncid, int varid, const char *name)
       if (!var)
 	 return NC_ENOTVAR;
       if (var->created)
-      {
 	 locid = var->hdf_datasetid;
-      }
    }
 
    /* Now find the attribute by name or number. */
-   for (att = *attlist; att; att = att->next)
+   for (att = *attlist; att; att = att->l.next)
       if (!strcmp(att->name, name))
 	 break;
 
    /* If att is NULL, we couldn't find the attribute. */
-   if (!att) 
+   if (!att)
       BAIL_QUIET(NC_ENOTATT);
-   
+
    /* Delete it from the HDF5 file, if it's been created. */
    if (att->created)
+   {
+      assert(locid);
+
       if(H5Adelete(locid, att->name) < 0)
 	 BAIL(NC_EATTMETA);
+   }
 
    /* Renumber all following attributes. */
-   for (natt = att->next; natt; natt = natt->next)
+   for (natt = att->l.next; natt; natt = natt->l.next)
       natt->attnum--;
 
    /* Delete this attribute from this list. */
    if ((retval = nc4_att_list_del(attlist, att)))
       BAIL(retval);
-   
+
  exit:
    if (datasetid > 0) H5Dclose(datasetid);
    return retval;
 }
 
 /* Write an attribute with type conversion. */
-int
-nc4_put_att_tc(int ncid, int varid, const char *name, nc_type file_type, 
-	       nc_type mem_type, int mem_type_is_long, size_t len, 
+static int
+nc4_put_att_tc(int ncid, int varid, const char *name, nc_type file_type,
+	       nc_type mem_type, int mem_type_is_long, size_t len,
 	       const void *op)
 {
    NC *nc;
@@ -786,7 +835,7 @@ nc4_put_att_tc(int ncid, int varid, const char *name, nc_type file_type,
 
    /* The length needs to be positive (cast needed for braindead
       systems with signed size_t). */
-   if((unsigned long) len > X_INT_MAX) 
+   if((unsigned long) len > X_INT_MAX)
       return NC_EINVAL;
 
    /* Find metadata. */
@@ -797,35 +846,36 @@ nc4_put_att_tc(int ncid, int varid, const char *name, nc_type file_type,
    h5 = NC4_DATA(nc);
    assert(h5);
 
-#ifdef USE_PNETCDF
+#if 0 /*def USE_PNETCDF*/
    /* Take care of files created/opened with parallel-netcdf library. */
    if (h5->pnetcdf_file)
    {
       if (mem_type == NC_UBYTE)
 	 mem_type = NC_BYTE;
+
       switch(mem_type)
       {
 	 case NC_BYTE:
-	    return ncmpi_put_att_schar(nc->int_ncid, varid, name, 
+	    return ncmpi_put_att_schar(nc->int_ncid, varid, name,
 				     file_type, len, op);
 	 case NC_CHAR:
-	    return ncmpi_put_att_text(nc->int_ncid, varid, name, 
+	    return ncmpi_put_att_text(nc->int_ncid, varid, name,
 				    len, op);
 	 case NC_SHORT:
-	    return ncmpi_put_att_short(nc->int_ncid, varid, name, 
+	    return ncmpi_put_att_short(nc->int_ncid, varid, name,
 				     file_type, len, op);
 	 case NC_INT:
 	    if (mem_type_is_long)
-	       return ncmpi_put_att_long(nc->int_ncid, varid, name, 
+	       return ncmpi_put_att_long(nc->int_ncid, varid, name,
 				       file_type, len, op);
 	    else
-	       return ncmpi_put_att_int(nc->int_ncid, varid, name, 
+	       return ncmpi_put_att_int(nc->int_ncid, varid, name,
 				      file_type, len, op);
 	 case NC_FLOAT:
-	    return ncmpi_put_att_float(nc->int_ncid, varid, name, 
+	    return ncmpi_put_att_float(nc->int_ncid, varid, name,
 				     file_type, len, op);
 	 case NC_DOUBLE:
-	    return ncmpi_put_att_double(nc->int_ncid, varid, name, 
+	    return ncmpi_put_att_double(nc->int_ncid, varid, name,
 				      file_type, len, op);
 	 case NC_NAT:
 	 default:
@@ -835,20 +885,20 @@ nc4_put_att_tc(int ncid, int varid, const char *name, nc_type file_type,
 #endif /* USE_PNETCDF */
 
    /* Otherwise, handle things the netcdf-4 way. */
-   return nc4_put_att(ncid, nc, varid, name, file_type, mem_type, len, 
+   return nc4_put_att(ncid, nc, varid, name, file_type, mem_type, len,
 		      mem_type_is_long, op);
 }
 
 /* Read an attribute of any type, with type conversion. This may be
  * called by any of the nc_get_att_* functions. */
 int
-nc4_get_att_tc(int ncid, int varid, const char *name, nc_type mem_type, 
+nc4_get_att_tc(int ncid, int varid, const char *name, nc_type mem_type,
 	       int mem_type_is_long, void *ip)
 {
    NC *nc;
    NC_HDF5_FILE_INFO_T *h5;
 
-   LOG((3, "nc4_get_att_tc: ncid 0x%x varid %d name %s mem_type %d", 
+   LOG((3, "nc4_get_att_tc: ncid 0x%x varid %d name %s mem_type %d",
 	ncid, varid, name, mem_type));
 
    /* Find metadata. */
@@ -859,12 +909,13 @@ nc4_get_att_tc(int ncid, int varid, const char *name, nc_type mem_type,
    h5 = NC4_DATA(nc);
    assert(h5);
 
-#ifdef USE_PNETCDF
+#if 0 /*def USE_PNETCDF*/
    /* Take care of files created/opened with parallel-netcdf library. */
    if (h5->pnetcdf_file)
    {
       if (mem_type == NC_UBYTE)
 	 mem_type = NC_BYTE;
+
       switch(mem_type)
       {
 	 case NC_BYTE:
@@ -889,12 +940,12 @@ nc4_get_att_tc(int ncid, int varid, const char *name, nc_type mem_type,
    }
 #endif /* USE_PNETCDF */
 
-   return nc4_get_att(ncid, nc, varid, name, NULL, mem_type, 
+   return nc4_get_att(ncid, nc, varid, name, NULL, mem_type,
 		      NULL, NULL, mem_type_is_long, ip);
 }
 
 int
-NC4_put_att(int ncid, int varid, const char *name, nc_type xtype, 
+NC4_put_att(int ncid, int varid, const char *name, nc_type xtype,
 	           size_t nelems, const void *value, nc_type memtype)
 {
    return nc4_put_att_tc(ncid, varid, name, xtype, memtype, 0, nelems, value);
